import require$$0 from 'https';
import require$$1 from 'http';
import require$$2 from 'url';
import { Agent as Agent$1 } from 'node:http';
import { Agent } from 'node:https';

function getDefaultExportFromCjs (x) {
	return x && x.__esModule && Object.prototype.hasOwnProperty.call(x, 'default') ? x['default'] : x;
}

const https = require$$0;
const http = require$$1;
const { URL: URL$1 } = require$$2;

let HttpProxyAgent$1 = class HttpProxyAgent extends http.Agent {
  constructor (options) {
    const { proxy, proxyRequestOptions, ...opts } = options;
    super(opts);
    this.proxy = typeof proxy === 'string'
      ? new URL$1(proxy)
      : proxy;
    this.proxyRequestOptions = proxyRequestOptions || {};
  }

  createConnection (options, callback) {
    const requestOptions = {
      ...this.proxyRequestOptions,
      method: 'CONNECT',
      host: this.proxy.hostname,
      port: this.proxy.port,
      path: `${options.host}:${options.port}`,
      setHost: false,
      headers: { ...this.proxyRequestOptions.headers, connection: this.keepAlive ? 'keep-alive' : 'close', host: `${options.host}:${options.port}` },
      agent: false,
      timeout: options.timeout || 0
    };

    if (this.proxy.username || this.proxy.password) {
      const base64 = Buffer.from(`${decodeURIComponent(this.proxy.username || '')}:${decodeURIComponent(this.proxy.password || '')}`).toString('base64');
      requestOptions.headers['proxy-authorization'] = `Basic ${base64}`;
    }

    if (this.proxy.protocol === 'https:') {
      requestOptions.servername = this.proxy.hostname;
    }

    const request = (this.proxy.protocol === 'http:' ? http : https).request(requestOptions);
    request.once('connect', (response, socket, head) => {
      request.removeAllListeners();
      socket.removeAllListeners();
      if (response.statusCode === 200) {
        callback(null, socket);
      } else {
        socket.destroy();
        callback(new Error(`Bad response: ${response.statusCode}`), null);
      }
    });

    request.once('timeout', () => {
      request.destroy(new Error('Proxy timeout'));
    });

    request.once('error', err => {
      request.removeAllListeners();
      callback(err, null);
    });

    request.end();
  }
};

let HttpsProxyAgent$1 = class HttpsProxyAgent extends https.Agent {
  constructor (options) {
    const { proxy, proxyRequestOptions, ...opts } = options;
    super(opts);
    this.proxy = typeof proxy === 'string'
      ? new URL$1(proxy)
      : proxy;
    this.proxyRequestOptions = proxyRequestOptions || {};
  }

  createConnection (options, callback) {
    const requestOptions = {
      ...this.proxyRequestOptions,
      method: 'CONNECT',
      host: this.proxy.hostname,
      port: this.proxy.port,
      path: `${options.host}:${options.port}`,
      setHost: false,
      headers: { ...this.proxyRequestOptions.headers, connection: this.keepAlive ? 'keep-alive' : 'close', host: `${options.host}:${options.port}` },
      agent: false,
      timeout: options.timeout || 0
    };

    if (this.proxy.username || this.proxy.password) {
      const base64 = Buffer.from(`${decodeURIComponent(this.proxy.username || '')}:${decodeURIComponent(this.proxy.password || '')}`).toString('base64');
      requestOptions.headers['proxy-authorization'] = `Basic ${base64}`;
    }

    // Necessary for the TLS check with the proxy to succeed.
    if (this.proxy.protocol === 'https:') {
      requestOptions.servername = this.proxy.hostname;
    }

    const request = (this.proxy.protocol === 'http:' ? http : https).request(requestOptions);
    request.once('connect', (response, socket, head) => {
      request.removeAllListeners();
      socket.removeAllListeners();
      if (response.statusCode === 200) {
        const secureSocket = super.createConnection({ ...options, socket });
        callback(null, secureSocket);
      } else {
        socket.destroy();
        callback(new Error(`Bad response: ${response.statusCode}`), null);
      }
    });

    request.once('timeout', () => {
      request.destroy(new Error('Proxy timeout'));
    });

    request.once('error', err => {
      request.removeAllListeners();
      callback(err, null);
    });

    request.end();
  }
};

var hpagent = {
  HttpProxyAgent: HttpProxyAgent$1,
  HttpsProxyAgent: HttpsProxyAgent$1
};

const mod = /*@__PURE__*/getDefaultExportFromCjs(hpagent);

const HttpProxyAgent = mod.HttpProxyAgent;
const HttpsProxyAgent = mod.HttpsProxyAgent;

class QuickLRU extends Map {
	#size = 0;
	#cache = new Map();
	#oldCache = new Map();
	#maxSize;
	#maxAge;
	#onEviction;

	constructor(options = {}) {
		super();

		if (!(options.maxSize && options.maxSize > 0)) {
			throw new TypeError('`maxSize` must be a number greater than 0');
		}

		if (typeof options.maxAge === 'number' && options.maxAge === 0) {
			throw new TypeError('`maxAge` must be a number greater than 0');
		}

		this.#maxSize = options.maxSize;
		this.#maxAge = options.maxAge || Number.POSITIVE_INFINITY;
		this.#onEviction = options.onEviction;
	}

	// For tests.
	get __oldCache() {
		return this.#oldCache;
	}

	#emitEvictions(cache) {
		if (typeof this.#onEviction !== 'function') {
			return;
		}

		for (const [key, item] of cache) {
			this.#onEviction(key, item.value);
		}
	}

	#deleteIfExpired(key, item) {
		if (typeof item.expiry === 'number' && item.expiry <= Date.now()) {
			if (typeof this.#onEviction === 'function') {
				this.#onEviction(key, item.value);
			}

			return this.delete(key);
		}

		return false;
	}

	#getOrDeleteIfExpired(key, item) {
		const deleted = this.#deleteIfExpired(key, item);
		if (deleted === false) {
			return item.value;
		}
	}

	#getItemValue(key, item) {
		return item.expiry ? this.#getOrDeleteIfExpired(key, item) : item.value;
	}

	#peek(key, cache) {
		const item = cache.get(key);

		return this.#getItemValue(key, item);
	}

	#set(key, value) {
		this.#cache.set(key, value);
		this.#size++;

		if (this.#size >= this.#maxSize) {
			this.#size = 0;
			this.#emitEvictions(this.#oldCache);
			this.#oldCache = this.#cache;
			this.#cache = new Map();
		}
	}

	#moveToRecent(key, item) {
		this.#oldCache.delete(key);
		this.#set(key, item);
	}

	* #entriesAscending() {
		for (const item of this.#oldCache) {
			const [key, value] = item;
			if (!this.#cache.has(key)) {
				const deleted = this.#deleteIfExpired(key, value);
				if (deleted === false) {
					yield item;
				}
			}
		}

		for (const item of this.#cache) {
			const [key, value] = item;
			const deleted = this.#deleteIfExpired(key, value);
			if (deleted === false) {
				yield item;
			}
		}
	}

	get(key) {
		if (this.#cache.has(key)) {
			const item = this.#cache.get(key);
			return this.#getItemValue(key, item);
		}

		if (this.#oldCache.has(key)) {
			const item = this.#oldCache.get(key);
			if (this.#deleteIfExpired(key, item) === false) {
				this.#moveToRecent(key, item);
				return item.value;
			}
		}
	}

	set(key, value, {maxAge = this.#maxAge} = {}) {
		const expiry = typeof maxAge === 'number' && maxAge !== Number.POSITIVE_INFINITY
			? (Date.now() + maxAge)
			: undefined;

		if (this.#cache.has(key)) {
			this.#cache.set(key, {
				value,
				expiry,
			});
		} else {
			this.#set(key, {value, expiry});
		}

		return this;
	}

	has(key) {
		if (this.#cache.has(key)) {
			return !this.#deleteIfExpired(key, this.#cache.get(key));
		}

		if (this.#oldCache.has(key)) {
			return !this.#deleteIfExpired(key, this.#oldCache.get(key));
		}

		return false;
	}

	peek(key) {
		if (this.#cache.has(key)) {
			return this.#peek(key, this.#cache);
		}

		if (this.#oldCache.has(key)) {
			return this.#peek(key, this.#oldCache);
		}
	}

	delete(key) {
		const deleted = this.#cache.delete(key);
		if (deleted) {
			this.#size--;
		}

		return this.#oldCache.delete(key) || deleted;
	}

	clear() {
		this.#cache.clear();
		this.#oldCache.clear();
		this.#size = 0;
	}

	resize(newSize) {
		if (!(newSize && newSize > 0)) {
			throw new TypeError('`maxSize` must be a number greater than 0');
		}

		const items = [...this.#entriesAscending()];
		const removeCount = items.length - newSize;
		if (removeCount < 0) {
			this.#cache = new Map(items);
			this.#oldCache = new Map();
			this.#size = items.length;
		} else {
			if (removeCount > 0) {
				this.#emitEvictions(items.slice(0, removeCount));
			}

			this.#oldCache = new Map(items.slice(removeCount));
			this.#cache = new Map();
			this.#size = 0;
		}

		this.#maxSize = newSize;
	}

	* keys() {
		for (const [key] of this) {
			yield key;
		}
	}

	* values() {
		for (const [, value] of this) {
			yield value;
		}
	}

	* [Symbol.iterator]() {
		for (const item of this.#cache) {
			const [key, value] = item;
			const deleted = this.#deleteIfExpired(key, value);
			if (deleted === false) {
				yield [key, value.value];
			}
		}

		for (const item of this.#oldCache) {
			const [key, value] = item;
			if (!this.#cache.has(key)) {
				const deleted = this.#deleteIfExpired(key, value);
				if (deleted === false) {
					yield [key, value.value];
				}
			}
		}
	}

	* entriesDescending() {
		let items = [...this.#cache];
		for (let i = items.length - 1; i >= 0; --i) {
			const item = items[i];
			const [key, value] = item;
			const deleted = this.#deleteIfExpired(key, value);
			if (deleted === false) {
				yield [key, value.value];
			}
		}

		items = [...this.#oldCache];
		for (let i = items.length - 1; i >= 0; --i) {
			const item = items[i];
			const [key, value] = item;
			if (!this.#cache.has(key)) {
				const deleted = this.#deleteIfExpired(key, value);
				if (deleted === false) {
					yield [key, value.value];
				}
			}
		}
	}

	* entriesAscending() {
		for (const [key, value] of this.#entriesAscending()) {
			yield [key, value.value];
		}
	}

	get size() {
		if (!this.#size) {
			return this.#oldCache.size;
		}

		let oldCacheSize = 0;
		for (const key of this.#oldCache.keys()) {
			if (!this.#cache.has(key)) {
				oldCacheSize++;
			}
		}

		return Math.min(this.#size + oldCacheSize, this.#maxSize);
	}

	get maxSize() {
		return this.#maxSize;
	}

	entries() {
		return this.entriesAscending();
	}

	forEach(callbackFunction, thisArgument = this) {
		for (const [key, value] of this.entriesAscending()) {
			callbackFunction.call(thisArgument, value, key, this);
		}
	}

	get [Symbol.toStringTag]() {
		return JSON.stringify([...this.entriesAscending()]);
	}
}

var parseUrl = require$$2.parse;

var DEFAULT_PORTS = {
  ftp: 21,
  gopher: 70,
  http: 80,
  https: 443,
  ws: 80,
  wss: 443,
};

var stringEndsWith = String.prototype.endsWith || function(s) {
  return s.length <= this.length &&
    this.indexOf(s, this.length - s.length) !== -1;
};

/**
 * @param {string|object} url - The URL, or the result from url.parse.
 * @return {string} The URL of the proxy that should handle the request to the
 *  given URL. If no proxy is set, this will be an empty string.
 */
function getProxyForUrl(url) {
  var parsedUrl = typeof url === 'string' ? parseUrl(url) : url || {};
  var proto = parsedUrl.protocol;
  var hostname = parsedUrl.host;
  var port = parsedUrl.port;
  if (typeof hostname !== 'string' || !hostname || typeof proto !== 'string') {
    return '';  // Don't proxy URLs without a valid scheme or host.
  }

  proto = proto.split(':', 1)[0];
  // Stripping ports in this way instead of using parsedUrl.hostname to make
  // sure that the brackets around IPv6 addresses are kept.
  hostname = hostname.replace(/:\d*$/, '');
  port = parseInt(port) || DEFAULT_PORTS[proto] || 0;
  if (!shouldProxy(hostname, port)) {
    return '';  // Don't proxy URLs that match NO_PROXY.
  }

  var proxy =
    getEnv('npm_config_' + proto + '_proxy') ||
    getEnv(proto + '_proxy') ||
    getEnv('npm_config_proxy') ||
    getEnv('all_proxy');
  if (proxy && proxy.indexOf('://') === -1) {
    // Missing scheme in proxy, default to the requested URL's scheme.
    proxy = proto + '://' + proxy;
  }
  return proxy;
}

/**
 * Determines whether a given URL should be proxied.
 *
 * @param {string} hostname - The host name of the URL.
 * @param {number} port - The effective port of the URL.
 * @returns {boolean} Whether the given URL should be proxied.
 * @private
 */
function shouldProxy(hostname, port) {
  var NO_PROXY =
    (getEnv('npm_config_no_proxy') || getEnv('no_proxy')).toLowerCase();
  if (!NO_PROXY) {
    return true;  // Always proxy if NO_PROXY is not set.
  }
  if (NO_PROXY === '*') {
    return false;  // Never proxy if wildcard is set.
  }

  return NO_PROXY.split(/[,\s]/).every(function(proxy) {
    if (!proxy) {
      return true;  // Skip zero-length hosts.
    }
    var parsedProxy = proxy.match(/^(.+):(\d+)$/);
    var parsedProxyHostname = parsedProxy ? parsedProxy[1] : proxy;
    var parsedProxyPort = parsedProxy ? parseInt(parsedProxy[2]) : 0;
    if (parsedProxyPort && parsedProxyPort !== port) {
      return true;  // Skip if ports don't match.
    }

    if (!/^[.*]/.test(parsedProxyHostname)) {
      // No wildcards, so stop proxying if there is an exact match.
      return hostname !== parsedProxyHostname;
    }

    if (parsedProxyHostname.charAt(0) === '*') {
      // Remove leading wildcard.
      parsedProxyHostname = parsedProxyHostname.slice(1);
    }
    // Stop proxying if the hostname ends with the no_proxy host.
    return !stringEndsWith.call(hostname, parsedProxyHostname);
  });
}

/**
 * Get the value for an environment variable.
 *
 * @param {string} key - The name of the environment variable.
 * @return {string} The value of the environment variable.
 * @private
 */
function getEnv(key) {
  return process.env[key.toLowerCase()] || process.env[key.toUpperCase()] || '';
}

var getProxyForUrl_1 = getProxyForUrl;

const defaultAgentOpts = {
  maxSockets: 64,
  keepAlive: false
};
class TimeoutError extends Error {
  constructor(message) {
    super(message);
    this.name = "TimeoutError";
    Error.captureStackTrace?.(this, TimeoutError);
  }
}
const inputToUrl = (url) => url instanceof URL ? url : new URL(url);
const inputToStr = (url) => url instanceof URL ? String(url) : url;
function fetchEnhanced(fetchImplementation, { undici = false, agentCacheSize = 512 } = { undici: false }) {
  const agentCache = new QuickLRU({ maxSize: agentCacheSize });
  async function getAgent(url, agentOpts = {}) {
    const { origin, protocol } = inputToUrl(url);
    const proxyUrl = agentOpts?.noProxy ? null : getProxyForUrl_1(url instanceof URL ? String(url) : url);
    const agentCacheKey = JSON.stringify({ proxyUrl, origin, ...agentOpts });
    if (agentCache.has(agentCacheKey)) return agentCache.get(agentCacheKey);
    let agent;
    if ("noProxy" in agentOpts) delete agentOpts.noProxy;
    if (undici) {
      const undiciOpts = { ...agentOpts };
      if ("keepAlive" in undiciOpts && !("pipelining" in undiciOpts)) {
        undiciOpts.pipelining = undiciOpts.keepAlive ? 1 : 0;
      }
      if ("keepAlive" in undiciOpts) {
        delete undiciOpts.keepAlive;
      }
      if ("maxSockets" in undiciOpts && typeof undiciOpts.maxSockets === "number") {
        undiciOpts.connections = undiciOpts.maxSockets;
        delete undiciOpts.maxSockets;
      }
      let UndiciProxyAgent;
      let UndiciAgent;
      let hadError = false;
      try {
        ({ ProxyAgent: UndiciProxyAgent, Agent: UndiciAgent } = await import('undici'));
      } catch {
        hadError = true;
      }
      if (proxyUrl && hadError) {
        throw new Error(`Please install the "undici" package to enable proxy support`);
      }
      if (proxyUrl && UndiciProxyAgent) {
        agent = new UndiciProxyAgent({ ...undiciOpts, uri: proxyUrl });
      } else if (UndiciAgent) {
        agent = new UndiciAgent(undiciOpts);
      }
    } else {
      const isHttps = protocol === "https:";
      if (proxyUrl) {
        agent = new (isHttps ? HttpsProxyAgent : HttpProxyAgent)({ ...agentOpts, proxy: proxyUrl });
      } else {
        agent = new (isHttps ? Agent : Agent$1)(agentOpts);
      }
    }
    agentCache.set(agentCacheKey, agent);
    return agent;
  }
  const fetch = (url, { timeout = 0, agentOpts = {}, ...opts } = {}) => {
    return new Promise(async (resolve, reject) => {
      if (!undici && !("agent" in opts)) {
        const agent = await getAgent(url, { ...defaultAgentOpts, ...agentOpts });
        if (agent) opts.agent = agent;
      } else if (undici && !("dispatcher" in opts)) {
        const agent = await getAgent(url, { ...defaultAgentOpts, ...agentOpts });
        if (agent) opts.dispatcher = agent;
      }
      let timeoutId;
      let controller;
      if (timeout) {
        if (!("signal" in opts) && globalThis.AbortController) {
          controller = new AbortController();
          opts.signal = controller.signal;
        }
        timeoutId = setTimeout(() => {
          controller?.abort?.();
          const err = new TimeoutError(`${opts.method || "GET"} ${inputToStr(url)} timed out after ${timeout}ms`);
          reject(err);
        }, timeout);
        timeoutId?.unref?.();
      }
      fetchImplementation(url, opts).then((res) => {
        if (timeoutId) clearTimeout(timeoutId);
        resolve(res);
      }).catch((err) => {
        if (timeoutId) clearTimeout(timeoutId);
        if (err.name === "AbortError") return resolve(new Response());
        reject(err);
      });
    });
  };
  fetch.clearCache = () => {
    for (const agent of agentCache.values()) {
      if ("destroy" in agent) agent.destroy();
    }
    agentCache.clear();
  };
  return fetch;
}

export { TimeoutError, fetchEnhanced as default };
